﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Gravemind
{
    public partial class DecoratorEditor : UserControl
    {
        public Device Device;
        public Material m;
        public Decorator Deco;
        public VertexBuffer vb;
        bool Pause = false;
        public Map map;

        public DecoratorEditor(Map map)
        {
            InitializeComponent();
            this.map = map;
            Deco = new Decorator(map);
        }

        private void DecoratorEditor_Load(object sender, EventArgs e)
        {
            Initialize();
        }

        public void Initialize()
        {
            panel1.Refresh();
            using (DecoratorEditor form = (DecoratorEditor)this)
            {
                if (!form.InitializeGraphics())
                {
                    MessageBox.Show("Could not initialize Direct3D!");
                }
                else
                {
                    form.Show();
                    form.Focus();
                    while (form.Created)
                    {
                        if (form.Focused)
                        {
                            form.Render();
                            try { Application.DoEvents(); }
                            catch { }
                        }
                        //else
                        //{
                        //    break;
                        //}
                    }
                }
            }
        }

        public void Update(Map map)
        {
            this.map = map;
            Deco = new Decorator(map);
            Initialize();
        }

        public bool InitializeGraphics()
        {
            PresentParameters parameters = new PresentParameters();
            parameters.Windowed = true;
            parameters.SwapEffect = SwapEffect.Discard;
            parameters.AutoDepthStencilFormat = DepthFormat.D16;
            parameters.EnableAutoDepthStencil = true;
            parameters.PresentationInterval = PresentInterval.One;
            IEnumerator enumerator = Manager.Adapters.GetEnumerator();
            while (enumerator.MoveNext())
            {
                AdapterInformation current = enumerator.Current as AdapterInformation;
                int adapter = current.Adapter;
                Caps deviceCaps = Manager.GetDeviceCaps(adapter, DeviceType.Hardware);
                CreateFlags softwareVertexProcessing = CreateFlags.SoftwareVertexProcessing;
                if (deviceCaps.DeviceCaps.SupportsHardwareTransformAndLight)
                {
                    softwareVertexProcessing = CreateFlags.HardwareVertexProcessing;
                }
                Device = new Device(adapter, deviceCaps.DeviceType, panel1, softwareVertexProcessing, new PresentParameters[] { parameters });
                if (Device != null)
                {
                    break;
                }
            }
            Device.DeviceReset += new EventHandler(OnResetDevice);
            Pause = false;
            OnResetDevice(Device, null);
            OnCreateDevice(Device, null);
            m = new Material();
            m.Diffuse = Color.Snow;
            m.Ambient = Color.Snow;
            m.Specular = Color.Snow;
            return true;
        }

        public void OnCreateDevice(object sender, EventArgs e)
        {
            Device device = (Device)sender;
            vb = new VertexBuffer(typeof(CustomVertex.PositionNormalTextured), 100, device, Usage.WriteOnly, VertexFormats.PositionNormal, Pool.Default);
            vb.Created += new EventHandler(OnCreateVertexBuffer);
            OnCreateVertexBuffer(vb, null);
        }

        public void OnCreateVertexBuffer(object sender, EventArgs e)
        {
            VertexBuffer buffer = (VertexBuffer)sender;
            CustomVertex.PositionNormalTextured[] texturedArray = (CustomVertex.PositionNormalTextured[])buffer.Lock(0, LockFlags.None);
            if (Deco.DataSize != 0)
            {
                for (int i = 0; i <= Deco.Verts.Verts.Count; i++)
                {
                    try
                    {
                        texturedArray[i].X = Deco.Verts.Verts[i].X;
                        texturedArray[i].Y = Deco.Verts.Verts[i].Y;
                        texturedArray[i].Z = Deco.Verts.Verts[i].Z;
                        texturedArray[i].Nx = Deco.Verts.Verts[i].NormalI;
                        texturedArray[i].Ny = Deco.Verts.Verts[i].NormalJ;
                        texturedArray[i].Nz = Deco.Verts.Verts[i].NormalK;
                        texturedArray[i].Tu = Deco.Verts.Verts[i].U;
                        texturedArray[i].Tv = Deco.Verts.Verts[i].V;
                    }
                    catch
                    {
                        break;
                    }
                }
            }
            buffer.Unlock();
        }

        private void OnPaint(object sender, PaintEventArgs e)
        {
            Render();
        }

        private void OnResetDevice(object sender, EventArgs e)
        {
            Device device = (Device)sender;
            device.RenderState.CullMode = Cull.None;
            device.RenderState.Lighting = true;
            device.RenderState.ZBufferEnable = true;
            device.RenderState.FillMode = FillMode.WireFrame;
            if (device.DeviceCaps.SourceBlendCaps.SupportsInverseSourceAlpha && device.DeviceCaps.SourceBlendCaps.SupportsSourceAlpha)
            {
                Device.RenderState.SourceBlend = Blend.SourceAlpha;
                Device.RenderState.DestinationBlend = Blend.InvSourceAlpha;
            }
            if (device.DeviceCaps.TextureFilterCaps.SupportsMinifyLinear)
            {
                Device.SamplerState[0].MinFilter = TextureFilter.Linear;
                Device.SamplerState[1].MinFilter = TextureFilter.Linear;
                Device.SamplerState[2].MinFilter = TextureFilter.Linear;
                Device.SamplerState[3].MinFilter = TextureFilter.Linear;
            }
            if (device.DeviceCaps.TextureFilterCaps.SupportsMagnifyLinear)
            {
                Device.SamplerState[0].MagFilter = TextureFilter.Linear;
                Device.SamplerState[1].MagFilter = TextureFilter.Linear;
                Device.SamplerState[2].MagFilter = TextureFilter.Linear;
                Device.SamplerState[3].MagFilter = TextureFilter.Linear;
            }
            if (device.DeviceCaps.TextureFilterCaps.SupportsMipMapLinear)
            {
                Device.SamplerState[0].MipFilter = TextureFilter.Linear;
                Device.SamplerState[1].MipFilter = TextureFilter.Linear;
                Device.SamplerState[2].MipFilter = TextureFilter.Linear;
                Device.SamplerState[3].MipFilter = TextureFilter.Linear;
            }
            Device.RenderState.Ambient = Color.White;
        }

        private void Render()
        {
            if (!Pause)
            {
                Device.Clear(ClearFlags.ZBuffer | ClearFlags.Target, Color.Blue, 1f, 0);
                Device.BeginScene();
                SetupMatrices();
                Device.Material = m;
                Device.RenderState.FillMode = FillMode.Solid;
                Texture Text = new Texture(Device, Deco.Shader.Bitmap, Usage.SoftwareProcessing, Pool.Managed);
                Device.SetTexture(0, Text);
                Device.SetStreamSource(0, vb, 0);
                Device.VertexFormat = VertexFormats.PositionNormal;
                Device.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 0x62);
                Device.RenderState.AlphaTestEnable = false;
                Device.RenderState.AlphaBlendEnable = false;
                Device.EndScene();
                Device.Present();
            }
        }

        private void SetupMatrices()
        {
            Device.Transform.World = Microsoft.DirectX.Matrix.RotationY(((float)Environment.TickCount) / 3000f);
            Device.Transform.View = Microsoft.DirectX.Matrix.LookAtLH(new Microsoft.DirectX.Vector3(0f, 3f, -5f), new Microsoft.DirectX.Vector3(0f, 0f, 0f), new Microsoft.DirectX.Vector3(0f, 1f, 0f));
            Device.Transform.Projection = Microsoft.DirectX.Matrix.PerspectiveFovLH(0.2f, 1f, 1f, 100f);
        }
    }
}
