﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Drawing;

namespace HaloMap
{
    public enum MapType
    {
        Singleplayer = 0,
        Multiplayer = 1,
        Mainmenu = 2,
        Shared = 3,
        SingleplayerShared = 4
    }

    public enum RawLocation : long 
    {
        Internal = 0X00000000,
        Mainmenu = 0XC0000000,
        Shared = 0X80000000,
        SingleplayerShared = 0X40000000,
        Unknown
    }

    public class Map
    {
        /// <summary>
        /// File Stream
        /// </summary>
        public FileStream fs;
        /// <summary>
        /// Reader
        /// </summary>
        public BinaryReader br;
        /// <summary>
        /// Writer
        /// </summary>
        public BinaryWriter bw;
        /// <summary>
        /// Tags
        /// </summary>
        public List<Tag> Tags = new List<Tag>();
        /// <summary>
        /// Script Refs
        /// </summary>
        public List<ScriptReference> ScriptReferences = new List<ScriptReference>();
        /// <summary>
        /// Selected Tag
        /// </summary>
        public Tag SelectedTag;
        /// <summary>
        /// Bsp List
        /// </summary>
        public List<Bsp> Bsps = new List<Bsp>();
        /// <summary>
        /// Lightmap List
        /// </summary>
        public List<Lightmap> Lightmaps = new List<Lightmap>();
        /// <summary>
        /// Status Label
        /// </summary>
        public ToolStripStatusLabel lb;
        /// <summary>
        /// Header
        /// </summary>
        public Header Header = new Header();
        /// <summary>
        /// Index
        /// </summary>
        public Index Index = new Index();
        /// <summary>
        /// Tags
        /// </summary>
        public Tags TagHandler = new Tags();
        /// <summary>
        /// Scripts
        /// </summary>
        public Scripts Scripts = new Scripts();
        /// <summary>
        /// Unicode Tables
        /// </summary>
        public Unicode Unicode = new Unicode();

        public Map(string Location)
        {
            fs = new FileStream(Location, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite);
            br = new BinaryReader(fs);
            bw = new BinaryWriter(fs);
        }

        public void Intialize(ToolStripStatusLabel lb, TreeView TV)
        {
            this.lb = lb;
            UpdateProccess("Reading Header...");
            Header.Read(this);
            UpdateProccess("Reading Index...");
            Index.Read(this);
            UpdateProccess("Reading Tags...");
            if (Header.MapType == MapType.Multiplayer)
            {
                TagHandler.ReadMP(this, TV);
            }
            else
            {
                TagHandler.Read(this, TV);
            }
            UpdateProccess("Reading Script Data...");
            Scripts.ReadReferences(this);
            if (Header.MapType != MapType.Multiplayer)
            {
                UpdateProccess("Reading Unicode Tables...");
                Unicode.Read(this);
            }
            if (VerifySettings())
            {
                UpdateProccess("Reading Coconuts...");
                Coconuts.Read(this);
            }
            UpdateProccess("Reading Lightmaps...");
            ReadLightmaps();
            UpdateProccess("Reading Bsp's...");
            ReadBsps();
            //PairBSPLightmap();
            UpdateProccess("Idle...");
        }

        public Image LoadPreview()
        {
            Tag scnr;
            TagReference Bitm = new TagReference();
            Reflexive Description = new Reflexive();
            Reflexive Level = new Reflexive();
            if (Header.MapType != MapType.Multiplayer)
            {
                scnr = Tags[3];
                br.BaseStream.Position = scnr.MetaOffset + 920;
                Description.Read(this, true);
                br.BaseStream.Position = Description.Translation + 8;
                Level.Read(this, true);
                br.BaseStream.Position = Level.Translation + 8;
                Bitm.Read(this, true);
                SelectedTag = Bitm.Tag;
                BitmapRaw.Read(this);
                SelectedTag = null;
            }
            else
            {
                scnr = Tags[0];
                br.BaseStream.Position = scnr.MetaOffset + 920;
                Description.Read(this, true);
                br.BaseStream.Position = Description.Translation + 16;
                Level.Read(this, true);
                br.BaseStream.Position = Level.Translation + 4;
                Bitm.Read(this, true);
                SelectedTag = Bitm.Tag;
                BitmapRaw.Read(this);
                SelectedTag = null;
            }
            return Bitmap.LoadPreview(this, BitmapRaw.RawChunks[0]);
        }

        public void Clear()
        {
            Header = new Header();
            Index = new Index();
            Tags.Clear();
            TagHandler = new Tags();
            Bsps.Clear();
            Lightmaps.Clear();
            Scripts = new Scripts();
            Unicode = new Unicode();
            ScriptReferences.Clear();
            SelectedTag = null;
        }

        public void Dispose()
        {
            br.Close();
            bw.Close();
            fs.Dispose();
            lb.Dispose();
            Header = null;
            Index = null;
            TagHandler = null;
            Scripts = null;
            Unicode = null;
            ScriptReferences = null;
            Bsps = null;
            Lightmaps = null;
            SelectedTag = null;
            Tags = null;
            Coconuts.Dispose();
            GC.Collect();
        }

        public void UpdateProccess(string Text)
        {
            lb.Text = Text;
            Application.DoEvents();
        }

        public void ReadBsps()
        {
            foreach (Tag t in Tags)
            {
                if (t.Class == "sbsp")
                {
                    Bsp bsp = new Bsp();
                    bsp.Read(this, t);
                    Bsps.Add(bsp);
                }
            }
        }
        public void ReadLightmaps()
        {
            foreach (Tag t in Tags)
            {
                if (t.Class == "ltmp")
                {
                    Lightmap Lightmap = new Lightmap();
                    Lightmap.Read(this, t);
                    Lightmaps.Add(Lightmap);
                }
            }
        }
        public void PairBSPLightmap()
        {
            for (int i = 0; i < Bsps.Count; i++)
            {
                try { Bsps[i].Lightmap = Lightmaps[i]; }
                catch { MessageBox.Show("Warning: BSP #" + i.ToString() + " does not have a Lightmap!"); }
            }
        }

        public int FindBsp(int Ident)
        {
            int Index = -1;
            foreach (Bsp bsp in Bsps)
            {
                if (bsp.tag.Identifier == Ident)
                {
                    return Index;
                }
                Index++;
            }
            return -1;
        }
        public int FindLightmap(int Ident)
        {
            int Index = -1;
            foreach (Lightmap lm in Lightmaps)
            {
                if (lm.Tag.Identifier == Ident)
                {
                    return Index;
                }
                Index++;
            }
            return -1;
        }

        public Tag FindTag(string Class, string Path)
        {
            foreach (Tag t in Tags)
            {
                if (t.Class == Class && t.Path == Path)
                {
                    return t;
                }
            }
            Tag tt = new Tag();
            tt.Class = Class;
            tt.Identifier = -1;
            tt.Path = "Null";
            return tt;
        }

        public Tag FindTag(int Ident)
        {
            foreach (Tag t in Tags)
            {
                if (t.Identifier == Ident)
                {
                    return t;
                }
            }
            Tag tt = new Tag();
            tt.Identifier = -1;
            tt.Path = "Null";
            return tt;
        }

        public int FindIdent(string Class, string Path)
        {
            foreach (Tag t in Tags)
            {
                if (t.Class == Class && t.Path == Path)
                {
                    return t.Identifier;
                }
            }
            return -1;
        }

        public int FindIndex(int Ident)
        {
            for (int i = 0; i < Tags.Count; i++)
            {
                if (Tags[i].Identifier == Ident)
                {
                    return i;
                }
            }
            return -1;
        }

        public int FindScript(string Script)
        {
            for (int i = 0; i < Header.SIDCount; i++)
            {
                if (ScriptReferences[i].Reference == Script)
                {
                    return i;
                }
            }
            return 0;
        }

        //public Tag GetSky()
        //{

        //}

        public int Padding()
        {
            return 512 - (Header.FileTableSize % 512);
        }

        public bool VerifySettings()
        {
            Globals.Initialize();
            if (File.Exists(Globals.Shared) == false)
            {
                MessageBox.Show("Error finding shared.map!");
                return false;
            }
            else if (File.Exists(Globals.MainMenu) == false)
            {
                MessageBox.Show("Error finding mainmenu.map!");
                return false;
            }
            else if (File.Exists(Globals.SingleplayerShared) == false)
            {
                MessageBox.Show("Error finding single_player_shared.map!");
                return false;
            }
            return true;
        }

        public BinaryReader OpenMap(RawLocation Type)
        {
            string Location = "";
            if (Type == RawLocation.Shared)
            {
                Location = Globals.Shared;
            }
            else if (Type == RawLocation.Mainmenu)
            {
                Location = Globals.MainMenu;
            }
            else if (Type == RawLocation.SingleplayerShared)
            {
                Location = Globals.SingleplayerShared;
            }
            else if (Type == RawLocation.Internal)
            {
                return br;
            }
            BinaryReader nbr = new BinaryReader(new FileStream(Location, FileMode.Open, FileAccess.Read, FileShare.Read));
            return nbr;
        }

        public void DecompileMap(string Location)
        {
            //Create a stream for our project file
            StreamWriter project = new StreamWriter(Location + "\\" + Header.MapName.Replace("\0", "") + ".haloproject");
            List<string> Decompiled = new List<string>();
            UpdateProccess("Intializing...");

            //Write every tag
            StreamWriter t = new StreamWriter(Location + "\\Tags.list");
            for (int i = 0; i < Index.TagCount; i++)
            {
                t.WriteLine(Tags[i].Path + "." + Tags[i].Class);
            }
            t.Close();

            //Write every StringId
            StreamWriter s = new StreamWriter(Location + "\\Strings.list");
            for (int i = 0; i < Index.TagCount; i++)
            {
                s.WriteLine(ScriptReferences[i].Reference);
            }
            s.Close();

            //Write All Unicode
            string[] Lang = { "English", "Japanese", "German", "French", "Spanish", "Italian", "Korean", "Chinese", "Portuguese" };
            for (int x = 0; x < 8; x++)
            {
                StreamWriter u = new StreamWriter(Location + "\\Unicode\\" + Lang[x] + ".unicode");
                for (int i = 0; i < Unicode.Tables[x].UnicodeStrings.Count; i++)
                {
                    u.WriteLine(Unicode.Tables[x].UnicodeStrings[i].Unicode);
                }
                u.Close();
            }

            //Loop and Decompile everything
            for (int i = 0; i < Index.TagCount; i++)
            {
                if (!Decompiled.Contains(Tags[i].Path + "." + Tags[i].Class))
                {
                    //Save Tag
                    UpdateProccess("Saving Tag (" + i.ToString() + "/" + Index.TagCount.ToString() + ")...");
                    SaveRecrusive(project, Tags[i], Location);
                }
            }

            //Clean Up
            UpdateProccess("Cleaning up Resources...");
            project.Close();
            project.Dispose();
            t.Dispose();
            s.Dispose();
            UpdateProccess("Idle...");
        }

        private void SaveRecrusive(StreamWriter sw, Tag Tag, string Location)
        {
            //Create Sub-Directorys
            int Pointer = Tag.Path.LastIndexOf("\\") + 1;
            string SubPath = "\\" + Tag.Path.Substring(0, Pointer);
            Directory.CreateDirectory(Location + SubPath);

            //Name of file
            string TempName = Tag.Path.Substring(Pointer, Tag.Path.Length - Pointer);
            string TempTagName = SubPath + TempName; // +"." + Tag.Class;
            TempTagName = TempTagName.Replace('*', '!');
            TempTagName = TempTagName.Replace("<", "_");
            TempTagName = TempTagName.Replace(">", "_");
            //TempTagName = TempTagName.Replace('/', '_');
            //TempTagName = TempTagName.Replace('\\', '_');

            //Save the Tag
            TagHandler.MorphMeta(this, Tag, Location + TempTagName);

            //Save Raw
            TagHandler.SaveRaw(this, Tag, Location);

            ////Recrusivly save idents
            //for (int i = 0; i < Tag.Idents.Count; i++)
            //{
            //    if (Tag.Idents[i].Tag.Identifier != -1)
            //    {
            //        SaveRecrusive(sw, Tag.Idents[i].Tag, Location);

            //        //Save Raw Data
            //        TagHandler.SaveRaw(this, Tag.Idents[i].Tag, Location);
            //    }
            //}

            //Add it to the list
            sw.WriteLine(TempTagName + "." + Tag.Class);
        }

        public void DecompileBSP(string Location)
        {
            //First Extract and fix the BSP tag
            Bsp Bsp = Bsps[0];
            this.SelectedTag = Bsp.tag;

            //Read Plugin
            PluginReader.ReadReferences(this);

            //Null References
            for (int i = 0; i < Bsp.tag.Idents.Count; i++)
            {
                Bsp.tag.Idents[i].Null(this);
            }

            //Null SIDs
            for (int i = 0; i < Bsp.tag.ScriptReferences.Count; i++)
            {
                Bsp.tag.ScriptReferences[i].Index = 0;
                Bsp.tag.ScriptReferences[i].Reference = "";
                Bsp.tag.ScriptReferences[i].Write(this);
            }

            //Fix Reflexives and move up 16 bytes
            for (int i = Bsp.tag.Reflexives.Count - 1; i > -1; i--)
            {
                if (Bsp.tag.Reflexives[i].Offset > 588)
                {
                    //Write Padding
                    bw.BaseStream.Position = Bsp.tag.MetaOffset + Bsp.tag.Reflexives[i].Offset;
                    bw.Write(new byte[16]);

                    //Write Reflexive Info
                    Bsp.tag.Reflexives[i].Offset += 16;
                    Bsp.tag.Reflexives[i].RawTranslation = (Bsp.tag.Reflexives[i].Translation - Bsp.tag.MetaOffset) + 16;
                    Bsp.tag.Reflexives[i].Write(this);
                }
                else
                {
                    //Write Reflexive Info
                    Bsp.tag.Reflexives[i].RawTranslation = (Bsp.tag.Reflexives[i].Translation - Bsp.tag.MetaOffset) + 16;
                    Bsp.tag.Reflexives[i].Write(this);
                }
            }

            //Create Memory Stream
            MemoryStream ms = new MemoryStream();
            ms.Write(new byte[16], 0, 16);
            br.BaseStream.Position = Bsp.tag.MetaOffset;
            ms.Write(br.ReadBytes(Bsp.tag.MetaSize), 0, Bsp.tag.MetaSize);

            //Write the New Tag
            File.WriteAllBytes(Location + "\\" + Header.MapName.Remove(Header.MapName.IndexOf('\0')) + ".bsp", ms.ToArray());

            //Raw------------------------------------------------------------------------------------------------

            //Detail Objects
            for (int i = 0; i <
        }

        public void Resign()
        {
            int size = Header.MapSize - 2048;
            int times = size / 4;
            int result = 0;
            int x;

            br.BaseStream.Position = 2048;
            for (x = 0; x < times; x++)
            {
                result ^= br.ReadInt32();
            }

            bw.BaseStream.Position = 752;
            bw.Write(result);
            Header.Checksum = result;
        }
    }
}
