﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;
using System.Windows.Forms;

namespace HaloMap
{
    public class ScriptReference
    {
        #region Fields
        public int Offset;
        public string Description;
        public string Reference = "";
        public int Index;
        public byte Length;
        #endregion

        #region Methods
        public void Read(Map map, int Offset)
        {
            this.Offset = Offset;
            Read(map);
        }
        public void Read(Map map)
        {
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + Offset;
            Index = map.br.ReadInt16();
            map.br.ReadByte();
            Length = map.br.ReadByte();
            try { Reference = map.ScriptReferences[Index].Reference; }
            catch (Exception) { }
        }
        public void Read(Map map, int Offset, bool b)
        {
            map.br.BaseStream.Position = Offset;
            Read(map, b);
        }
        public void Read(Map map, bool b)
        {
            Offset = (int)map.br.BaseStream.Position;
            Index = map.br.ReadInt16();
            map.br.ReadByte();
            Length = map.br.ReadByte();
            try { Reference = map.ScriptReferences[Index].Reference; }
            catch (Exception) { }
        }

        public void Write(Map map, int Offset)
        {
            this.Offset = Offset;
            Write(map);
        }
        public void Write(Map map)
        {
            map.bw.BaseStream.Position = map.SelectedTag.MetaOffset + Offset;
            map.bw.Write((short)Index);
            map.bw.Write((byte)0);
            map.bw.Write((byte)Reference.Length);
        }
        public void Write(Map map, bool b)
        {
            map.bw.BaseStream.Position = Offset;
            map.bw.Write((short)Index);
            map.bw.Write((byte)0);
            map.bw.Write((byte)Reference.Length);
        }
        #endregion
    }

    public class Scripts
    {
        public void ReadReferences(Map map)
        {
            //Index
            map.br.BaseStream.Position = map.Header.SIDIndexOffset;
            for (int i = 0; i < map.Header.SIDCount; i++)
            {
                ScriptReference S = new ScriptReference();
                S.Offset = map.br.ReadInt32();
                S.Index = i;
                map.ScriptReferences.Add(S);
            }

            //FileTable
            map.br.BaseStream.Position = map.Header.SIDTableOffset;
            foreach (ScriptReference S in map.ScriptReferences)
            {
                char chr;
                while ((chr = map.br.ReadChar()) != '\0')
                    S.Reference += chr;
            }
        }

        public void AddScriptReference(Map map, string Reference)
        {
            //Build a ScriptReference
            ScriptReference S = new ScriptReference();
            S.Reference = Reference;
            S.Index = map.Header.SIDCount;
            S.Offset = 128 * map.Header.SIDCount;
            map.ScriptReferences.Add(S);

            //Lets set some stuff up
            byte[] EOF;
            int Shift = 0;

            //Add it to the 128 Table
            map.UpdateProccess("Adding Script Reference...");
            map.br.BaseStream.Position = map.Header.SIDMetaTableOffset + (128 * map.Header.SIDCount);
            EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
            map.bw.BaseStream.Position = map.Header.SIDMetaTableOffset + (128 * map.Header.SIDCount);
            map.bw.Write(S.Reference.ToCharArray());
            for (int i = S.Reference.Length; i < 128; i++)
            {
                map.bw.Write((byte)0);
            }
            map.bw.Write(EOF);
            Shift += 128;

            //Shift
            map.Header.SIDIndexOffset += Shift;

            //Add it to the Index
            map.br.BaseStream.Position = map.Header.SIDIndexOffset + (map.Header.SIDCount * 4);
            EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
            map.bw.BaseStream.Position = map.Header.SIDIndexOffset + (map.Header.SIDCount * 4);
            map.bw.Write(map.Header.SIDTableSize);
            map.bw.Write(EOF);
            Shift += 4;

            //Shift
            map.Header.SIDTableOffset += Shift;

            //Add it to the String Table
            map.br.BaseStream.Position = map.Header.SIDTableOffset + map.Header.SIDTableSize;
            EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
            map.bw.BaseStream.Position = map.Header.SIDTableOffset + map.Header.SIDTableSize;
            map.bw.Write(S.Reference.ToCharArray());
            map.bw.Write((byte)0);
            map.bw.Write(EOF);
            Shift += S.Reference.Length + 1;

            //Shift
            map.UpdateProccess("Shifting...");
            map.Header.SIDTableSize += S.Reference.Length + 1;
            map.Header.FileTableIndexOffset += Shift;
            map.Header.FileTableOffset += Shift;
            map.Header.IndexOffset += Shift;
            map.bw.BaseStream.Position = 8;
            map.bw.Write((int)map.bw.BaseStream.Length);
            map.bw.BaseStream.Position = 16;
            map.bw.Write(map.Header.IndexOffset);
            map.bw.BaseStream.Position = 368;
            map.bw.Write(map.Header.SIDCount);
            map.bw.Write(map.Header.SIDTableSize);
            map.bw.Write(map.Header.SIDIndexOffset);
            map.bw.Write(map.Header.SIDTableOffset);
            map.bw.BaseStream.Position = 720;
            map.bw.Write(map.Header.FileTableOffset);
            map.bw.BaseStream.Position += 4;
            map.bw.Write(map.Header.FileTableIndexOffset);

            //Bitmaps
            map.UpdateProccess("Refreshing Info...");
            map.Header.Read(map);
            map.Index.Read(map);
            map.TagHandler.Read(map);
            map.UpdateProccess("Fixing Raw Pointers...");
            for (int i = 0; i < map.Index.TagCount; i++)
            {
                if (map.Tags[i].Class == "bitm")
                {
                    map.SelectedTag = map.Tags[i];
                    Reflexive BitmapData = new Reflexive();
                    BitmapData.Read(map, map.Tags[i].MetaOffset + 68);
                    for (int Chunk = 0; Chunk < BitmapData.ChunkCount; Chunk++)
                    {
                        map.br.BaseStream.Position = BitmapData.Translation + map.SelectedTag.MetaOffset + (Chunk * 116);
                        map.br.BaseStream.Position += 28;
                        int Offset1 = map.br.ReadInt32();
                        int Offset2 = map.br.ReadInt32();
                        int Offset3 = map.br.ReadInt32();
                        if (Offset1 != 0 | Offset1 != -1 && RawData.RawLocator(Offset1) == RawLocation.Internal) { Offset1 += Shift; }
                        if (Offset2 != 0 | Offset2 != -1 && RawData.RawLocator(Offset1) == RawLocation.Internal) { Offset2 += Shift; }
                        if (Offset3 != 0 | Offset3 != -1 && RawData.RawLocator(Offset1) == RawLocation.Internal) { Offset3 += Shift; }
                        map.bw.BaseStream.Position = BitmapData.Translation + map.SelectedTag.MetaOffset + (Chunk * 116);
                        map.bw.BaseStream.Position += 28;
                        map.bw.Write(Offset1);
                        map.bw.Write(Offset2);
                        map.bw.Write(Offset3);
                    }
                }
            }

            //Unicode
            map.UpdateProccess("Shifting Unicode Tables...");
            map.br.BaseStream.Position = map.Tags[0].MetaOffset + 400;
            for (int i = 0; i < 8; i++)
            {
                int Offset = (int)map.br.BaseStream.Position;
                int StringCount = map.br.ReadInt32();
                int TableSize = map.br.ReadInt32();
                int IndexOffset = map.br.ReadInt32() + Shift;
                int TableOffset = map.br.ReadInt32() + Shift;
                map.bw.BaseStream.Position = Offset + 8;
                map.bw.Write(IndexOffset);
                map.bw.Write(TableOffset);
                map.br.BaseStream.Position += 8;
            }

            map.UpdateProccess("Idle...");
        }
    }
}
