﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace HaloMap
{
    public class UnicodeEntry
    {
        #region Fields
        private string unicode;
        /// <summary>
        /// Unicode String
        /// </summary>
        public string Unicode { get { return unicode; } set { unicode = value; } }

        private int length;
        /// <summary>
        /// Length
        /// </summary>
        public int Length { get { return length; } set { length = value; } }

        private int indexoffset;
        /// <summary>
        /// Index Offset
        /// </summary>
        public int IndexOffset { get { return indexoffset; } set { indexoffset = value; } }

        private int tableoffset;
        /// <summary>
        /// Table Offset
        /// </summary>
        public int TableOffset { get { return tableoffset; } set { tableoffset = value; } }

        private short id;
        /// <summary>
        /// Identifier
        /// </summary>
        public short ID { get { return id; } set { id = value; } }

        private short unknown;
        /// <summary>
        /// Unknown
        /// </summary>
        public short Unknown { get { return unknown; } set { unknown = value; } }
        #endregion
    }

    public class UnicodeTable
    {
        #region Fields
        private int stringcount;
        /// <summary>
        /// Unicode String Count
        /// </summary>
        public int UnicodeStringCount { get { return stringcount; } set { stringcount = value; } }

        private int tablesize;
        /// <summary>
        /// Unicode Table Size
        /// </summary>
        public int TableSize { get { return tablesize; } set { tablesize = value; } }

        private int indexoffset;
        /// <summary>
        /// Unicode Table Index Offset
        /// </summary>
        public int IndexOffset { get { return indexoffset; } set { indexoffset = value; } }

        private int tableoffset;
        /// <summary>
        /// Unicode Table Offset
        /// </summary>
        public int TableOffset { get { return tableoffset; } set { tableoffset = value; } }

        private List<UnicodeEntry> unicodestrings = new List<UnicodeEntry>();
        /// <summary>
        /// Unicode Strings
        /// </summary>
        public List<UnicodeEntry> UnicodeStrings { get { return unicodestrings; } set { unicodestrings = value; } }
        #endregion
    }

    public class Unicode
    {
        #region Fields
        private int pointers;
        /// <summary>
        /// Offset to the Pointers
        /// </summary>
        public int UnicodePointers { get { return pointers; } set { pointers = value; } }

        private UnicodeTable[] tables = new UnicodeTable[9];
        /// <summary>
        /// Unicode Tables
        /// </summary>
        public UnicodeTable[] Tables { get { return tables; } set { tables = value; } }
        #endregion

        public Unicode()
        {
            Globals.Initialize();
            for (int i = 0; i < 9; i++)
            {
                Tables[i] = new UnicodeTable();
                Tables[i].UnicodeStrings = new List<UnicodeEntry>();
            }
        }

        public void Read(Map map)
        {
            map.br.BaseStream.Position = map.Tags[0].MetaOffset + 400;
            UnicodePointers = (int)map.br.BaseStream.Position;
            //Pointers
            for (int i = 0; i < 9; i++)
            {
                Tables[i].UnicodeStringCount = map.br.ReadInt32();
                Tables[i].TableSize = map.br.ReadInt32();
                Tables[i].IndexOffset = map.br.ReadInt32();
                Tables[i].TableOffset = map.br.ReadInt32();
                map.br.BaseStream.Position += 12;
            }

            //Tables
            for (int i = 0; i < 9; i++)
            {
                //Index
                map.br.BaseStream.Position = Tables[i].IndexOffset;
                for (int x = 0; x < Tables[i].UnicodeStringCount; x++)
                {
                    UnicodeEntry Uni = new UnicodeEntry();
                    Uni.ID = map.br.ReadInt16();
                    Uni.Unknown = map.br.ReadInt16();
                    Uni.TableOffset = map.br.ReadInt32();
                    Tables[i].UnicodeStrings.Add(Uni);
                }

                //Table
                map.br.BaseStream.Position = Tables[i].TableOffset;
                for (int x = 0; x < Tables[i].UnicodeStringCount; x++)
                {
                    while (map.br.BaseStream.Position < Tables[i].TableOffset + Tables[i].UnicodeStrings[x].TableOffset)
                    {
                        byte b = map.br.ReadByte();
                        switch (b)
                        {
                            case 226:
                                {
                                    short punc = map.br.ReadInt16();
                                    if (punc == -26240)
                                    {
                                        Tables[i].UnicodeStrings[x].Unicode += "'";
                                    }
                                    else
                                    {
                                        Tables[i].UnicodeStrings[x].Unicode += (char)b;
                                        map.br.BaseStream.Position -= 2;
                                    }
                                    Tables[i].UnicodeStrings[x].Length += 1;
                                    break;
                                }
                            case 238:
                                {
                                    short image = map.br.ReadInt16();
                                    int Index = Array.IndexOf(Globals.uniDwords, image);
                                    if (Index == -1) Tables[i].UnicodeStrings[x].Unicode += "[?]";
                                    else Tables[i].UnicodeStrings[x].Unicode += Globals.uniTranslations[Index];
                                    Tables[i].UnicodeStrings[x].Length += 4;
                                    break;
                                }
                            case 13:
                                {
                                    map.br.BaseStream.Position++;
                                    Tables[i].UnicodeStrings[x].Unicode += @" \r\n ";
                                    Tables[i].UnicodeStrings[x].Length += 1;
                                    break;
                                }
                            default:
                                {
                                    Tables[i].UnicodeStrings[x].Unicode += (char)b;
                                    Tables[i].UnicodeStrings[x].Length += 1;
                                    break;
                                }
                        }
                    }
                }
            }
        }
    }
}
