﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace HaloMap
{
    public partial class Bitmask : UserControl
    {
        public string Description;
        public object Value;
        public List<string> Items = new List<string>();
        public List<int> Values = new List<int>();
        public int Offset;
        public Type t;
        int size;

        public Bitmask(string Description, object Value, List<string> Items, List<int> Values, int Offset, Type t)
        {
            InitializeComponent();
            this.Description = Description;
            this.Value = Value;
            this.Items = Items;
            this.Values = Values;
            this.Offset = Offset;
            this.t = t;
            if (t == typeof(byte))
            {
                size = 8;
            }
            else if (t == typeof(short))
            {
                size = 16;
            }
            else if (t == typeof(int))
            {
                size = 32;
            }
        }

        private void Bitmask_Load(object sender, EventArgs e)
        {
            byte[] Bytes = new byte[0];
            string Val = Value.ToString();
            if (t == typeof(byte))
            {
                try { Bytes = BitConverter.GetBytes(Convert.ToByte(Val)); }
                catch { Bytes = BitConverter.GetBytes(byte.Parse("0")); }

            }
            else if (t == typeof(short))
            {
                Bytes = BitConverter.GetBytes(Convert.ToUInt16(Val));
            }
            else if (t == typeof(int))
            {
                Bytes = BitConverter.GetBytes(Convert.ToUInt32(Val));
            }
            bool[] Bits = ConvertToBoolArray(Bytes);
            groupPanel1.Text = Description;
            List<CheckBox> Boxes = new List<CheckBox>();
            for (int i = 0; i < size; i++)
            {
                CheckBox c = new CheckBox();
                if (i < Items.Count)
                {
                    c.Text = Items[i];
                    c.Tag = Values[i];
                }
                else
                {
                    c.Text = "Bit - " + i.ToString();
                    c.Tag = i;
                    c.Visible = false;
                }
                if (Bits[(int)c.Tag] == true)
                {
                    c.Checked = true;
                }
                Boxes.Add(c);
            }
            Boxes.Reverse();
            foreach (CheckBox c in Boxes)
            {
                groupPanel1.Controls.Add(c);
                c.Dock = DockStyle.Top;
            }
        }

        public void Update(object Value)
        {
            byte[] Bytes = new byte[0];
            string Val = Value.ToString();
            if (t == typeof(byte))
            {
                Bytes = BitConverter.GetBytes(Convert.ToByte(Val));

            }
            else if (t == typeof(short))
            {
                Bytes = BitConverter.GetBytes(Convert.ToUInt16(Val));
            }
            else if (t == typeof(int))
            {
                Bytes = BitConverter.GetBytes(Convert.ToUInt32(Val));
            }
            bool[] Bits = ConvertToBoolArray(Bytes);
            for (int i = 0; i < size; i++)
            {
                CheckBox c = (CheckBox)groupPanel1.Controls[i];
                if (Bits[(int)c.Tag] == true)
                {
                    c.Checked = true;
                }
                else
                {
                    c.Checked = false;
                }
            }
        }

        public uint Save()
        {
            List<bool> bits = new List<bool>();
            uint tmp = 1;
            uint tmpVal = 0;
            for (int x = groupPanel1.Controls.Count - 1; x > -1; x--)
            {
                CheckBox c = (CheckBox)groupPanel1.Controls[x];
                if (c.Checked) tmpVal += tmp;
                tmp <<= 1; //shift
            }
            return tmpVal;
        }

        public static bool[] ConvertToBoolArray(byte[] Bytes)
        {
            bool[] bitArray = new bool[Bytes.Length * 8];
            int input;
            int bittester;

            for (int i = 0; i < Bytes.Length; i++)
            {
                input = BitConverter.ToInt16(new byte[2] { Bytes[i], 0 }, 0);
                bittester = 1;

                for (int counter = 0; counter < 8; counter++)
                {
                    bitArray[(i * 8) + counter] = (input & bittester) > 0 ? true : false;
                    bittester *= 2;
                }
            }

            return bitArray;
        }

        public static uint ConvertToInt(List<bool> Bits)
        {
            uint intToShift = 1;
            uint tempIntToWrite = 0;
            for (int counter = 0; counter < Bits.Count; counter++)
            {
                if (Bits[counter] == true)
                {
                    tempIntToWrite += intToShift;
                }
                intToShift <<= 1;
            }
            return tempIntToWrite;
        }

        public static byte[] ConvertToByteArray(List<bool> Bits)
        {
            int BitData;
            BitData = 0;

            for (int z = 0; z < Bits.Count; z++)
            {
                BitData = setBit(BitData, z, Bits[z]);
            }

            byte[] AllBytes = BitConverter.GetBytes(BitData);
            byte[] ReturnBytes = new byte[Bits.Count / 8];

            for (int i = 0; i < Bits.Count / 8; i++)
            {
                ReturnBytes[i] = AllBytes[i];
            }

            return ReturnBytes;
        }

        private static int setBit(int integer, int bitIndex, bool onOff)
        {
            if (onOff == true)
            {
                integer = (1 << bitIndex) | integer;
            }
            else
            {
                integer = ~(1 << bitIndex) & integer;
            }
            return integer;
        }

        public void Poke(ProcessStream p, int Translation)
        {
            if (t == typeof(byte))
            {
                Value = p.ReadByte(Offset - Translation);
            }
            else if (t == typeof(short))
            {
                Value = p.ReadInt16(Offset - Translation);
            }
            else if (t == typeof(int))
            {
                Value = p.ReadInt32(Offset - Translation);
            }
            Update(Value);
        }
    }
}
