﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace HaloMap
{
    public partial class BSPViewer : Form
    {
        public Map map;
        public Bsp b;

        public Camera Cam;
        public Device Dev;
        public Material m;
        public VertexBuffer DetailObjects;

        public BSPViewer(Map map, Bsp b)
        {
            InitializeComponent();
            this.map = map;
            this.b = b;

            //Initialize
            Init();
        }

        private void BSPViewer_Load(object sender, EventArgs e)
        {
            
        }

        public void Init()
        {
            using (BSPViewer bsp = (BSPViewer)this)
            {
                //Initlialize Graphics
                bsp.InitGraphics();

                //Update Form
                Cam = new Camera(this);
                bsp.Show();
                bsp.Focus();

                //Render
                while (bsp.Created)
                {
                    try
                    {
                        Render();
                        Application.DoEvents();
                    }
                    catch { break; }
                }
            }
        }

        public void InitGraphics()
        {
            this.Show();
            this.Focus();
            PresentParameters PParms = new PresentParameters();
            PParms.Windowed = true;
            PParms.PresentationInterval = PresentInterval.Default;
            PParms.FullScreenRefreshRateInHz = PresentParameters.DefaultPresentRate;

            PParms.SwapEffect = SwapEffect.Discard;
            PParms.EnableAutoDepthStencil = true;
            PParms.AutoDepthStencilFormat = DepthFormat.D16;
            Dev = new Device(0, DeviceType.Hardware, this, CreateFlags.SoftwareVertexProcessing, PParms);
            Dev.DeviceReset += new EventHandler(OnResetDevice);
            OnResetDevice(Dev, null);
            SetUp();

            m = new Material();
            m.Ambient = Color.White;
            m.Diffuse = Color.White;
            m.Specular = Color.White;
        }

        public void Render()
        {
            //Clear Screen
            Dev.Clear(ClearFlags.ZBuffer | ClearFlags.Target, Color.Black, 1f, 0);

            //Begine Scene
            Dev.BeginScene();

            Cam.move();
            SetupMatrices();
            Dev.Material = m;
            Dev.RenderState.FillMode = FillMode.Solid;
            Dev.SetStreamSource(0, DetailObjects, 0);
            Dev.VertexFormat = VertexFormats.PositionNormal;
            Dev.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 0x62);
            Dev.RenderState.AlphaTestEnable = false;
            Dev.RenderState.AlphaBlendEnable = false;

            //Present
            Dev.EndScene();
            Dev.Present();
        }

        public void OnResetDevice(object sender, EventArgs e)
        {
            Device dev = (Device)sender;
            // Turn off culling, so we see the front and back of the triangle
            dev.RenderState.CullMode = Cull.None;
            // Turn off D3D lighting
            dev.RenderState.Lighting = true;
            // Turn on the ZBuffer
            dev.RenderState.ZBufferEnable = true;
            dev.RenderState.FillMode = FillMode.Solid;
            dev.RenderState.SourceBlend = Blend.SourceAlpha;
            dev.RenderState.DestinationBlend = Blend.InvSourceAlpha;
            dev.RenderState.Ambient = Color.White;
            Dev.SamplerState[0].MinFilter = TextureFilter.Linear;
            Dev.SamplerState[0].MagFilter = TextureFilter.Linear;
            Dev.SamplerState[0].MipFilter = TextureFilter.Linear;
            Dev.SamplerState[1].MinFilter = TextureFilter.Linear;
            Dev.SamplerState[1].MagFilter = TextureFilter.Linear;
            Dev.SamplerState[1].MipFilter = TextureFilter.Linear;
        }

        public void SetUp()
        {
            DetailObjects = new VertexBuffer(typeof(CustomVertex.PositionNormalTextured), 100, Dev, Usage.WriteOnly, VertexFormats.PositionNormal, Pool.Default);
            CustomVertex.PositionNormalTextured[] Data = (CustomVertex.PositionNormalTextured[])DetailObjects.Lock(0, LockFlags.None);
            for (int i = 0; i < b.DetailObjects.Count; i++)
            {
                try
                {
                    HaloMap.BSP.DetailObject Object = b.DetailObjects[i];
                    Data[i].X = Object.Raw.Vertices.Vertice[i].X;
                    Data[i].Y = Object.Raw.Vertices.Vertice[i].Y;
                    Data[i].Z = Object.Raw.Vertices.Vertice[i].Z;
                    Data[i].Nx = Object.Raw.Normals.Normals[i].X;
                    Data[i].Ny = Object.Raw.Normals.Normals[i].Y;
                    Data[i].Nz = Object.Raw.Normals.Normals[i].Z;
                    Data[i].Tu = Object.Raw.UVData.UVs[i].X;
                    Data[i].Tv = Object.Raw.UVData.UVs[i].Y;
                }
                catch { }
            }
            DetailObjects.Unlock();
        }

        public void SetupMatrices()
        {
            Dev.Transform.World = Matrix.Identity;
            Dev.Transform.View = Matrix.LookAtRH(Cam.Position, Cam.LookAt, Cam.UpVector);
            Dev.Transform.Projection = Matrix.PerspectiveFovRH((float)Math.PI / 4.0f, 1.0f, 0.1f, 10000.0f);
        }

        private void BSPViewer_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                Cam.oldx = e.X;
                Cam.oldy = e.Y;
            }
        }

        private void BSPViewer_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                Cam.change(e.X, e.Y);
            }
        }
    }
}
