﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;
using Microsoft.DirectX.Direct3D;
using Direct3D = Microsoft.DirectX.Direct3D;

namespace HaloMap
{
    public partial class CollisionViewer : Form
    {
        Microsoft.DirectX.Direct3D.Device Device;
        PresentParameters PParms = new PresentParameters();
        bool IsPaused = false;
        Collision Coll;
        Matrix WorldTransform = Matrix.Identity;
        Camera Cam;
        Material SphereMaterial;
        Material BlackMaterial;
        Material GreenMaterial;
        Material DefaultMaterial;
        bool Loaded = false;
        Map map;

        public CollisionViewer(Map map)
        {
            InitializeComponent();
            this.map = map;
            this.Text = "Collision Viewer - " + map.SelectedTag.Path;
            this.ClientSize = new System.Drawing.Size(800, 600);
            Intialize();
        }

        private void CollisionViewer_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                Cam.oldx = e.X;
                Cam.oldy = e.Y;
            }
        }

        private void CollisionViewer_MouseMove(object sender, MouseEventArgs e)
        {

            if (e.Button == MouseButtons.Right)
            {
                //Cam.change(e.X, e.Y);
            }
            //if (Cam.mousenavigation == Camera.MouseNavigationTypes.None)
            // {
            //    cam.StartMouseNavigation(e);
            //  }
            //  cam.NavigateByMouseTrackball(e);

        }

        protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
        {
            this.Render();
        }

        protected override void OnKeyPress(System.Windows.Forms.KeyPressEventArgs e)
        {
            if ((int)(byte)e.KeyChar == (int)System.Windows.Forms.Keys.Escape)
            {
                this.Dispose();
            }
        }

        protected override void OnResize(System.EventArgs e)
        {
            IsPaused = ((this.WindowState == FormWindowState.Minimized) || !this.Visible);
        }

        public void OnResetDevice(object sender, EventArgs e)
        {
            Direct3D.Device dev = (Direct3D.Device)sender;
            // Turn off culling, so we see the front and back of the triangle
            dev.RenderState.CullMode = Cull.None;
            // Turn off D3D lighting
            dev.RenderState.Lighting = true;
            // Turn on the ZBuffer
            dev.RenderState.ZBufferEnable = true;
            dev.RenderState.FillMode = FillMode.Solid;
            dev.RenderState.SourceBlend = Blend.SourceAlpha;
            dev.RenderState.DestinationBlend = Blend.InvSourceAlpha;
            dev.RenderState.Ambient = Color.White;
            Device.SamplerState[0].MinFilter = TextureFilter.Linear;
            Device.SamplerState[0].MagFilter = TextureFilter.Linear;
            Device.SamplerState[0].MipFilter = TextureFilter.Linear;
            Device.SamplerState[1].MinFilter = TextureFilter.Linear;
            Device.SamplerState[1].MagFilter = TextureFilter.Linear;
            Device.SamplerState[1].MipFilter = TextureFilter.Linear;
        }

        public void Intialize()
        {
            using (CollisionViewer frm = (CollisionViewer)this)
            {
                if (!frm.InitializeGraphics()) // Initialize Direct3D
                {
                    MessageBox.Show("Error could not initialize Direct3D!");
                    return;
                }
                Cam = new Camera(this);
                frm.Show();
                frm.Focus();
                try { Cam.AimCamera(Coll.Spheres[0].Position); }
                catch (Exception) { MessageBox.Show("0 Spheres, Could not focus camera!"); }
                Loaded = true;
                ReadSettings();
                while (frm.Created)
                {
                    frm.Render();
                    Application.DoEvents();
                }
            }
        }

        public bool InitializeGraphics()
        {
            this.Show();
            this.Focus();
            Application.DoEvents();
            PParms.Windowed = true; // We don't want to run fullscreen
            PParms.PresentationInterval = Direct3D.PresentInterval.Default;
            PParms.FullScreenRefreshRateInHz = Direct3D.PresentParameters.DefaultPresentRate;

            PParms.SwapEffect = SwapEffect.Discard; // Discard the frames 
            PParms.EnableAutoDepthStencil = true; // Turn on a Depth stencil
            PParms.AutoDepthStencilFormat = DepthFormat.D16; // And the stencil format
            Device = new Direct3D.Device(0,Direct3D.DeviceType.Hardware, this, CreateFlags.SoftwareVertexProcessing, PParms); //Create a device
            Device.DeviceReset += new System.EventHandler(this.OnResetDevice);
            this.OnResetDevice(Device, null);
            IsPaused = false;
            DefaultMaterial = new Material();
            DefaultMaterial.Diffuse = Color.White;
            DefaultMaterial.Ambient = Color.White;
            BlackMaterial = new Material();
            BlackMaterial.Diffuse = Color.Black;
            BlackMaterial.Ambient = Color.Black;
            GreenMaterial = new Material();
            GreenMaterial.Diffuse = Color.GreenYellow;
            GreenMaterial.Ambient = Color.GreenYellow;

            LoadMesh();
            return true;
        }

        public void LoadMesh()
        {
            Coll = new Collision(map);
            Coll.Intialize(map, ref Device);
        }

        public void Render()
        {
            if (!IsPaused)
            {
                Device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Globals.WorldColor, 1.0f, 0);
                Device.BeginScene();
                Cam.move();
                SetupMatrices();
                Device.RenderState.FillMode = Globals.FillMode;
                SphereMaterial = new Material();
                SphereMaterial.Diffuse = Globals.SphereColor;
                SphereMaterial.Ambient = Globals.SphereColor;
                Device.Material = DefaultMaterial;
                Device.Transform.World = Matrix.Identity;
                Device.Material = DefaultMaterial;
                Coll.DrawMeshes(ref Device);
                Device.Material = SphereMaterial;
                Coll.DrawSpheres(ref Device);
                Device.EndScene();
                Device.Present();
            }
        }

        public void SetupMatrices()
        {
            Device.Transform.World = Matrix.Identity;
            Device.Transform.View = Matrix.LookAtRH(Cam.Position, Cam.LookAt, Cam.UpVector);
            Device.Transform.Projection = Matrix.PerspectiveFovRH((float)Math.PI / 4.0f, 1.0f, 0.1f, 10000.0f);
        }

        private void CollisionViewer_FormClosing(object sender, FormClosingEventArgs e)
        {
            Cam.Dispose();
            Device = null;
            Coll = null;
            Dispose();
        }

        private void btnSphereColor_Click(object sender, EventArgs e)
        {
            ColorDialog cd = new ColorDialog();
            if (cd.ShowDialog() == DialogResult.OK)
            {
                Globals.SphereColor = cd.Color;
                btnSphereColor.BackColor = cd.Color;
                SaveSettings();
                Globals.Initialize();
                Render();
            }
        }

        public void SaveSettings()
        {
            StreamWriter sw = new StreamWriter(Application.StartupPath + "\\Settings\\Collision.cfg");
            sw.WriteLine(Globals.SphereColor.ToArgb());
            sw.WriteLine(Globals.WorldColor.ToArgb());
            sw.WriteLine(comboBox1.SelectedIndex.ToString());
            sw.Close();
        }

        private void btnWorldColor_Click(object sender, EventArgs e)
        {
            ColorDialog cd = new ColorDialog();
            if (cd.ShowDialog() == DialogResult.OK)
            {
                Globals.WorldColor = cd.Color;
                btnWorldColor.BackColor = cd.Color;
                SaveSettings();
                Globals.Initialize();
                Render();
            }
        }

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (Loaded)
            {
                SaveSettings();
                Globals.Initialize();
                Render();
            }
        }

        public void ReadSettings()
        {
            if (Loaded)
            {
                btnSphereColor.BackColor = Globals.SphereColor;
                btnWorldColor.BackColor = Globals.WorldColor;
                comboBox1.SelectedIndex = (int)Globals.FillMode;
            }
        }
    }
}
